class DimensionIndicesError(IndexError):
    """
    Error raises when dimension indices is out of range.
    """

    def __init__(self, index_name: str, value: int, max_value: int):
        """
        Instantiate a DimensionIndicesError.

        Args:
            index_name (str): name of the index.
            value (int): value of the indice.
            max_value (int): maximum value of the indice.
        """
        self.index_name = index_name
        self.value = value
        self.max_value = max_value

    def __str__(self):
        """
        Print the error.
        """
        return (
            f"`{self.index_name}` expect value between {-self.max_value-1} and"
            f" {self.max_value}, got {self.value} instead."
        )


class OutputNotComputedEror(RuntimeError):
    """
    Error raised if the visualizer hasn't a `ModelOutput` in theirs attributes.
    """

    def __str__(self):
        """
        Print the error.
        """
        return (
            "Visualizer must be called with `__call__` or `compute` method and a text"
            " input before applying `plot` method."
        )



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Optional, Union

import torch
from transformers import PreTrainedModel, PreTrainedTokenizer, PreTrainedTokenizerFast


class Visualizer(ABC):
    """
    Abstractclass for representing a Visualizer object.
    """

    @abstractmethod
    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Abtractmethod for instantiate a Visualizer object.

        Args:
            model (PreTrainedModel): A `model` given by HuggingFace ecosystem.
            tokenizer (Union[PreTrainedTokenizer, PreTrainedTokenizerFast]): A `tokenizer` given by HuggingFace ecosystem.
            device (Optional[Union[torch.device, str]]): A `device` where computation takes place. If no `torch.device` are provided, the `model`'s device is used.
        """
        self.model = deepcopy(model).eval()
        self.tokenizer = tokenizer

        if device is None:
            self._device = self.model.device
        else:
            self._device = device

    @property
    def device(self):
        """
        Return the `torch.device` used for computation.
        """
        return self._device

    @device.setter
    def device(self, value: Union[torch.device, str]) -> None:
        """
        Set a `torch.device` for computation.

        Args:
            value (Union[torch.device, str]): A `torch.device` or a `str`.

        Raises:
            ValueError: raised if `value` isn't a `torch.device` or acceptable `str`.
        """
        if isinstance(value, str):
            self._device = torch.device(value)
        elif isinstance(value, torch.device):
            self._device = value
        else:
            raise ValueError(
                "`value` expect a `torch.device` or a `str`, got"
                f" {type(value)} instead."
            )

        self.model = self.model.to(self._device)

    @abstractmethod
    def set_device(self, value: Union[torch.device, str]) -> None:
        """
        Set a `torch.device` for computation.

        Args:
            value (Union[torch.device, str]): A `torch.device` or a `str`.

        Raises:
            ValueError: raised if `value` isn't a `torch.device` or acceptable `str`.
        """
        self.device = value

    @abstractmethod
    def __call__(self, text: str) -> None:
        """
        Given a text input generates necessary elements for visualization. Multiple text input is not supported.

        Args:
            text (str): Text input

        Raises:
            NotImplementedError: If not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def compute(self, text: str):
        """
        Given a text input generates necessary elements for visualization. Multiple text input is not supported. Work in place.

        Args:
            text (str): Text input
        """
        self.__call__(text)
        return self

    @abstractmethod
    def plot(self, **plot_kwargs) -> None:
        """
        Plot the Visualizer. The purpose of kwargs are used to setup plotting parameters.

        Raises:
            NotImplementedError: If not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def __str__(self) -> str:
        """
        Return the class name, the model and the tokenizer of the Visualizer.

        Returns:
            str: Class name, model and tokenizer.
        """
        s = f"{self.__class__.__name__}("
        s += f"model={self.model.__class__.__name__}, "
        s += f"tokenizer={self.tokenizer.__class__.__name__})"
        return s



from typing import List, Optional, Set, Tuple, Union

import torch
from transformers import PreTrainedModel, PreTrainedTokenizer, PreTrainedTokenizerFast

from transformers_visualizer.errors import DimensionIndicesError, OutputNotComputedEror
from transformers_visualizer.plotting import plot_token_to_token_specific_dimension
from transformers_visualizer.visualizer import Visualizer


class TokenToTokenAttentions(Visualizer):
    """
    Visualizer for plotting token-to-token attentions matrices.
    """

    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Create a token-to-token attention matrices visualizer. Plot attention matrices given a model and a tokenizer.

        Args:
            model (PreTrainedModel): A `model` given by HuggingFace ecosystem.
            tokenizer (Union[PreTrainedTokenizer, PreTrainedTokenizerFast]): A `tokenizer` given by HuggingFace ecosystem.
            device (Optional[Union[torch.device, str]]): A `device` where computation takes place. If no `torch.device` are provided, the `model`'s device is used.
        """
        super().__init__(model, tokenizer, device)
        self.model.config.output_attentions = True

    def set_device(self, value: Union[torch.device, str]) -> None:
        """
        Set a `torch.device` for computation.

        Args:
            value (Union[torch.device, str]): A `torch.device` or a `str`.

        Raises:
            ValueError: raised if `value` isn't a `torch.device` or acceptable `str`.
        """
        super().set_device(value)

    def __call__(self, text: str) -> None:
        """
        Given a text input generates necessary elements for visualization. Multiple text input is not supported.

        Args:
            text (str): Text input.
        """
        if (
            isinstance(text, List) or isinstance(text, Tuple) or isinstance(text, Set)  # type: ignore
        ):
            raise NotImplementedError("Multiple text input is not supported.")

        self.tokens = self.tokenizer(text, return_tensors="pt").to(self._device)
        self.all_tokens: List[str] = self.tokenizer.convert_ids_to_tokens(
            self.tokens["input_ids"].squeeze().tolist()  # type: ignore
        )

        self.output = self.model(**self.tokens)
        self.attentions = torch.concat(self.output.attentions)

    def compute(self, text: str):
        """
        Given a text input generates necessary elements for visualization. Multiple text input is not supported. Work in place.

        Args:
            text (str): Text input.
        """
        return super().compute(text)

    def plot(self, layer_index: int = -1, **plot_kwargs) -> None:
        """
        Plot the Visualizer. The purpose of kwargs are used to setup plotting parameters.

        Args:
            layer_index (int): Layer index of the model to be plotted. Defaults to last layer, i.e. -1.
            figsize (Tuple[int, int], optional): Figsize of the plot. Defaults to (20, 20).
            ticks_fontsize (int, optional): Ticks fontsize. Defaults to 7.
            title_fontsize (int, optional): Title fontsize. Defaults to 9.
            cmap (str, optional): Colormap. Defaults to "viridis".
            colorbar (bool, optional): Display colorbars. Defaults to True.

        Raises:
            OutputNotComputedEror: raised if no `Output` present.
        """
        if not hasattr(self, "attentions") or not hasattr(self, "all_tokens"):
            raise OutputNotComputedEror

        max_layer_index = self.model.config.num_hidden_layers - 1
        try:
            attentions = self.attentions[layer_index]
        except IndexError:
            raise DimensionIndicesError("layer_index", layer_index, max_layer_index)

        plot_token_to_token_specific_dimension(
            attentions,
            self.all_tokens,
            "Head",
            **plot_kwargs,
        )

    def __str__(self) -> str:
        """
        Return the class name, the model and the tokenizer of the Visualizer.

        Returns:
            str: Class name, model and tokenizer.
        """
        return super().__str__()


class TokenToTokenNormalizedAttentions(Visualizer):
    """
    Visualizer for plotting token-to-token attentions matrices normalized on `attention_heads` dimension.
    """

    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        device: Optional[Union[torch.device, str]] = None,
    ):
        """
        Create a token-to-token normalized attention matrices visualizer. Plot normalized attention matrices across head axis given a model and a tokenizer.

        Args:
            model (PreTrainedModel): A `model` given by HuggingFace ecosystem.
            tokenizer (Union[PreTrainedTokenizer, PreTrainedTokenizerFast]): A `tokenizer` given by HuggingFace ecosystem.
            device (Optional[Union[torch.device, str]]): A `device` where computation takes place. If no `torch.device` are provided, the `model`'s device is used.
        """
        super().__init__(model, tokenizer, device)
        self.model.config.output_attentions = True

    def set_device(self, value: Union[torch.device, str]):
        """
        Set a `torch.device` for computation.

        Args:
            value (Union[torch.device, str]): A `torch.device` or a `str`.

        Raises:
            ValueError: raised if `value` isn't a `torch.device` or acceptable `str`.
        """
        return super().set_device(value)

    def __call__(self, text: str, order: Optional[Union[str, int]] = None) -> None:
        """
        Given a text input generates necessary elements for visualization. Multiple text input is not supported.

        Args:
            text (str): Text input.
            order (Optional[Union[str, int]]): Order used when `torch.norm` is applied.
        """
        tokens = self.tokenizer(text, return_tensors="pt").to(self._device)
        self.all_tokens: List[str] = self.tokenizer.convert_ids_to_tokens(
            tokens["input_ids"].squeeze().tolist()  # type: ignore
        )

        self.output = self.model(**tokens)
        self.attentions = torch.concat(self.output.attentions)
        self.normalized_attentions = torch.linalg.norm(
            self.attentions, dim=1, ord=order
        )

    def compute(self, text: str, order: Optional[Union[str, int]] = None):
        """
        Given a text input generates necessary elements for visualization. Multiple text input is not supported. Work in place.

        Args:
            text (str): Text input.
        """
        self.__call__(text, order)
        return self

    def plot(self, **plot_kwargs) -> None:
        """
        Plot the Visualizer. The purpose of kwargs are used to setup plotting parameters.

        Args:
            figsize (Tuple[int, int], optional): Figsize of the plot. Defaults to (20, 20).
            ticks_fontsize (int, optional): Ticks fontsize. Defaults to 7.
            title_fontsize (int, optional): Title fontsize. Defaults to 9.
            cmap (str, optional): Colormap. Defaults to "viridis".
            colorbar (bool, optional): Display colorbars. Defaults to True.

        Raises:
            OutputNotComputedEror: raised if no `Output` present.
        """
        if (
            not hasattr(self, "attentions")
            or not hasattr(self, "normalized_attentions")
            or not hasattr(self, "all_tokens")
        ):
            raise OutputNotComputedEror

        plot_token_to_token_specific_dimension(
            self.normalized_attentions,
            self.all_tokens,
            "Layer",
            **plot_kwargs,
        )

    def __str__(self) -> str:
        """
        Return the class name, the model and the tokenizer of the Visualizer.

        Returns:
            str: Class name, model and tokenizer.
        """
        return super().__str__()



# from copy import deepcopy
# from typing import List, Optional, Set, Tuple, Union

# import torch
# from captum.attr import LayerActivation
# from transformers import PreTrainedModel, PreTrainedTokenizer, PreTrainedTokenizerFast
# from transformers.tokenization_utils_base import BatchEncoding

# from transformers_visualizer.errors import OutputNotComputedEror
# from transformers_visualizer.plotting import plot_token_to_head_specific_dimension
# from transformers_visualizer.visualizer import Visualizer

# # TODO: interpretable_embedding  + gorokoba560/norm-analysis-of-transformer


# class TokenToHeadScore(Visualizer):
#     """
#     Visualizer for plotting token-to-head importance scores.
#     """

#     def __init__(
#         self,
#         model: PreTrainedModel,
#         tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
#         device: Optional[Union[torch.device, str]] = None,
#     ) -> None:
#         """
#         Create a token-to-head score visualizer. Visualize transformed vector (i.e. `hidden_state` in transformers-based models context).

#         Args:
#             model (PreTrainedModel): _description_
#             tokenizer (Union[PreTrainedTokenizer, PreTrainedTokenizerFast]): _description_
#             device (Optional[Union[torch.device, str]], optional): _description_. Defaults to None.
#         """
#         super().__init__(model, tokenizer, device)
#         self.model.config.output_attentions = True
#         self.model.config.output_hidden_states = True

#     def set_device(self, value: Union[torch.device, str]) -> None:
#         """
#         Set a `torch.device`.

#         Args:
#             value (Union[torch.device, str]): A `torch.device` or a `str`.

#         Raises:
#             ValueError: raised if `value` isn't a `torch.device` or acceptable `str`.
#         """
#         super().set_device(value)

#     def _forward(self, inputs, ids: BatchEncoding):
#         ids = deepcopy(ids)
#         ids.pop("input_ids", "None")
#         pred = self.model({**inputs, **ids.to(self.device)})
#         return pred.last_hidden_state

#     def __call__(self, text: str) -> None:
#         """
#         Given a text input generates necessary elements for visualization. Multiple text input is not supported.

#         Args:
#             text (str): Text input.
#         """
#         if (
#             isinstance(text, List) or isinstance(text, Tuple) or isinstance(text, Set)  # type: ignore
#         ):
#             raise NotImplementedError("Multiple text input is not supported.")

#         self.tokens = self.tokenizer(text, return_tensors="pt").to(self._device)
#         self.all_tokens: List[str] = self.tokenizer.convert_ids_to_tokens(
#             self.tokens["input_ids"].squeeze().tolist()  # type: ignore
#         )

#         self.output = self.model(**self.tokens)
#         self.attentions = torch.concat(self.output.attentions)

#         # n_layers = len(self.model.base_model.encoder.layer)  # type: ignore
#         n_layers = self.model.config.num_hidden_layers
#         n_heads = self.model.config.num_attention_heads
#         head_size = (
#             self.model.config.hidden_size // self.model.config.num_attention_heads
#         )
#         all_head_size = self.model.config.hidden_size

#         # Getting Access to Value Activations

#         layers = [
#             self.model.base_model.encoder.layer[layer].attention.self.value  # type: ignore
#             for layer in range(n_layers)
#         ]
#         input_embeddings = self.output.hidden_states[0]

#         la = LayerActivation(self._forward, layers)
#         value_layer_acts = la.attribute(
#             input_embeddings, additional_forward_args=(self.tokens)
#         )
#         # shape -> layer x seq_len x all_head_size
#         value_layer_acts = torch.concat(value_layer_acts)
#         new_x_shape = value_layer_acts.size()[:-1] + (n_heads, head_size)
#         value_layer_acts = value_layer_acts.view(*new_x_shape)

#         # layer x num_heads x 1 x head_size
#         value_layer_acts = (
#             value_layer_acts.permute(0, 1, 3, 2).permute(0, 1, 3, 2, 4).contiguous()
#         )
#         # layer x seq_length x num_heads x 1 x head_size
#         value_layer_acts = value_layer_acts.view(
#             value_layer_acts.size()[:-1]
#             + (
#                 1,
#                 value_layer_acts.size()[-1],
#             )
#         )

#         # Getting Access to Dense Features
#         dense_acts = torch.stack(
#             [
#                 dlayer.attention.output.dense.weight
#                 for dlayer in self.model.bert.encoder.layer  # type: ignore
#             ]
#         )
#         dense_acts = dense_acts.view(len(layers), all_head_size, n_heads, head_size)
#         # layer x num_heads x head_size x all_head_size
#         dense_acts = dense_acts.permute(0, 2, 3, 1).contiguous()

#         # layers, seq_length, num_heads, 1, all_head_size
#         f_x = torch.stack(
#             [
#                 value_layer_acts_i.matmul(dense_acts_i)
#                 for value_layer_acts_i, dense_acts_i in zip(
#                     value_layer_acts, dense_acts
#                 )
#             ]
#         )
#         # layer x seq_length x num_heads x 1 x all_head_size)
#         f_x = f_x.view(f_x.size()[:-2] + (f_x.size()[-1],))
#         f_x = f_x.permute(0, 1, 3, 2).contiguous()

#         self.f_x = f_x
#         self.f_x_normalized = torch.linalg.norm(self.f_x, dim=-1)

#     def compute(self, text: str):
#         """
#         Given a text input generates necessary elements for visualization. Multiple text input is not supported. Work in place.

#         Args:
#             text (str): Text input.
#         """
#         return super().compute(text)

#     def plot(self, **plot_kwargs) -> None:
#         """
#         Plot the Visualizer. The purpose of kwargs are used to setup plotting parameters.

#         Args:
#             figsize (Tuple[int, int], optional): Figsize of the plot. Defaults to (20, 20).
#             ticks_fontsize (int, optional): Ticks fontsize. Defaults to 7.
#             title_fontsize (int, optional): Title fontsize. Defaults to 9.
#             cmap (str, optional): Colormap. Defaults to "viridis".
#             colorbar (bool, optional): Display colorbars. Defaults to True.

#         Raises:
#             OutputNotComputedEror: raised if no `Output` present.
#         """
#         if (
#             not hasattr(self, "attentions")
#             or hasattr(self, "f_x_normalized")
#             or not hasattr(self, "all_tokens")
#         ):
#             raise OutputNotComputedEror

#         plot_token_to_head_specific_dimension(
#             self.attentions,
#             self.all_tokens,
#             "Layer",
#             **plot_kwargs,
#         )

#     def __str__(self) -> str:
#         return super().__str__()


# # def _visualize_token2head_scores(scores_mat, all_tokens):
# #     fig = plt.figure(figsize=(15, 15))

# #     for idx, scores in enumerate(scores_mat):
# #         scores_np = np.array(scores)
# #         ax = fig.add_subplot(6, 2, idx + 1)
# #         # append the attention weights
# #         im = ax.matshow(scores_np, cmap="viridis")

# #         fontdict = {"fontsize": 8}

# #         ax.set_xticks(range(len(all_tokens)))
# #         ax.set_yticks(range(len(scores)))

# #         ax.set_xticklabels(all_tokens, fontdict=fontdict, rotation=90)
# #         # ax.set_yticklabels(range(len(scores[0])), fontdict=fontdict)
# #         ax.set_yticklabels(range(len(scores)), fontdict=fontdict)
# #         ax.set_xlabel("Layer {}".format(idx + 1))

# #         fig.colorbar(im, fraction=0.046, pad=0.04)
# #     plt.tight_layout()
# #     return plt.gcf()


# # def _forward(inputs, model, ids, device):
# #     pred = model(
# #         ids["input_ids"].to(device),
# #         ids["attention_mask"].to(device))
# #     return pred.logits.max(1).values


# # def visualize_vector_norm(
# #     device, model, description, tokenizer, preprocessing_func=None, compute="norm", layer=None
# # ):
# #     if preprocessing_func is not None:
# #         description = preprocessing_func(description)
# #     all_tokens = tokenizer.convert_ids_to_tokens(tokenizer(description)["input_ids"])
# #     ids = tokenizer.encode_plus(
# #         description,
# #         truncation=True,
# #         padding=True,
# #         max_length=512,
# #         return_attention_mask=True,
# #         return_tensors="pt",
# #     )
# #     output = model(
# #         ids["input_ids"].to(device),
# #         ids["attention_mask"].to(device),
# #     )
# #     attentions = output.attentions
# #     output_attentions_all = torch.stack(attentions)

# #     output_attentions_all_shape = output_attentions_all.shape
# #     batch = output_attentions_all_shape[1]
# #     num_heads = output_attentions_all_shape[2]
# #     head_size = 64
# #     all_head_size = 768

# #     layers = [
# #         model.base_model.encoder.layer[layer].attention.self.value
# #         for layer in range(len(model.base_model.encoder.layer))
# #     ]

# #     input_embeddings = output.hidden_states[0]

# #     la = captum.attr.LayerActivation(_forward, layers)
# #     value_layer_acts = la.attribute(input_embeddings, additional_forward_args=(
# #         model,
# #         ids,
# #         device
# #     ))
# #     # shape -> layer x batch x seq_len x all_head_size
# #     value_layer_acts = torch.stack(value_layer_acts)

# #     new_x_shape = value_layer_acts.size()[:-1] + (num_heads, head_size)
# #     value_layer_acts = value_layer_acts.view(*new_x_shape)

# #     # layer x batch x neum_heads x 1 x head_size
# #     value_layer_acts = value_layer_acts.permute(0, 1, 3, 2, 4)

# #     value_layer_acts = value_layer_acts.permute(0, 1, 3, 2, 4).contiguous()
# #     value_layer_acts_shape = value_layer_acts.size()

# #     # layer x batch x seq_length x num_heads x 1 x head_size
# #     value_layer_acts = value_layer_acts.view(value_layer_acts_shape[:-1] + (1, value_layer_acts_shape[-1],))

# #     dense_acts = torch.stack([dlayer.attention.output.dense.weight for dlayer in model.base_model.encoder.layer])
# #     dense_acts = dense_acts.view(len(layers), all_head_size, num_heads, head_size)

# #     # layer x num_heads x head_size x all_head_size
# #     dense_acts = dense_acts.permute(0, 2, 3, 1).contiguous()

# #     # layers, batch, seq_length, num_heads, 1, all_head_size
# #     f_x = torch.stack([value_layer_acts_i.matmul(dense_acts_i) for value_layer_acts_i, dense_acts_i in zip(value_layer_acts, dense_acts)])

# #     # layer x batch x seq_length x num_heads x 1 x all_head_size)
# #     f_x_shape = f_x.size()
# #     f_x = f_x.view(f_x_shape[:-2] + (f_x_shape[-1],))
# #     f_x = f_x.permute(0, 1, 3, 2, 4).contiguous()

# #     #(layers x batch, num_heads, seq_length, all_head_size)
# #     f_x_shape = f_x.size()


# #     # ||f(x)||
# #     #(layers x batch, num_heads, seq_length)
# #     f_x_norm = torch.linalg.norm(f_x, dim=-1)
# #     if compute == "norm":
# #         _visualize_token2head_scores(f_x_norm.squeeze().detach().cpu().numpy(), all_tokens)


# #     # ||alpha * f(x)||
# #     # layer x batch x num_heads x seq_length x seq_length x all_head_size
# #     alpha_f_x = torch.einsum('lbhks,lbhsd->lbhksd', output_attentions_all, f_x)

# #     # layer x batch x num_heads x seq_length x seq_length
# #     alpha_f_x_norm = torch.linalg.norm(alpha_f_x, dim=-1)
# #     if compute == "alpha-norm":
# #         _visualize_token2token_scores(
# #             alpha_f_x_norm[layer].squeeze().detach().cpu().numpy(),
# #             all_tokens,
# #             "Head"
# #         )

# #     # || SUM alpha * f(x)||
# #     summed_alpha_f_x = alpha_f_x.sum(dim=2)

# #     # layers x batch x seq_length x seq_length
# #     summed_alpha_f_x_norm = torch.linalg.norm(summed_alpha_f_x, dim=-1)
# #     if compute == "sum-norm":
# #         _visualize_token2token_scores(
# #             summed_alpha_f_x_norm.squeeze().cpu().detach().numpy(),
# #             all_tokens,
# #             "Layer"
# #         )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['transformers_visualizer', 'transformers_visualizer.visualizers']

package_data = \
{'': ['*']}

install_requires = \
['captum>=0.5.0',
 'matplotlib>=3.5',
 'torchtyping>=0.1.4',
 'transformers>=4.0.0']

setup_kwargs = {
    'name': 'transformers-visualizer',
    'version': '0.2.0',
    'description': 'Explain your 🤗 transformers without effort! Display the internal behavior of your model.',
    'long_description': '<h1 align="center">Transformers visualizer</h1>\n<p align="center">Explain your 🤗 transformers without effort!</p>\n<h1 align="center"></h1>\n\n<p align="center">\n    <a href="https://opensource.org/licenses/Apache-2.0">\n        <img src="https://img.shields.io/badge/License-Apache%202.0-blue.svg"/>\n    </a>\n</p>\n\nTransformers visualizer is a python package designed to work with 🤗 [transformers](https://huggingface.co/docs/transformers/index) package. Given a <code>model</code> and a <code>tokenizer</code>, this package supports multiple ways to explain your model by plotting its internal behavior.\n\nThis package is mostly based on the [Captum][Captum] tutorials [[1]][captum_part1] [[2]][Captum_part2].\n\n## Installation\n\n```python\npip install transformers-visualizer\n```\n\n## Quickstart\n\nLet\'s define a model, a tokenizer and a text input for the following examples.\n\n```python\nfrom transformers import AutoModel, AutoTokenizer\n\nmodel_name = "bert-base-uncased"\nmodel = AutoModel.from_pretrained(model_name)\ntokenizer = AutoTokenizer.from_pretrained(model_name)\ntext = "The dominant sequence transduction models are based on complex recurrent or convolutional neural networks that include an encoder and a decoder."\n```\n\n<details><summary>Attention matrices</summary>\n<p align="center">Plot attention matrices of a specific layer</p>\n\n```python\nfrom transformers_visualizer import TokenToTokenAttentions\n\nvisualizer = TokenToTokenAttentions(model, tokenizer)\nvisualizer(text)\n```\n\nInstead of using `__call__` function, you can use the `compute` method. Both work in place, `compute` method allows chaining method.\n\n`plot` method accept a layer index as parameter to specify which part of your model you want to plot. By default, the last layer is plotted.\n\n```python\nimport matplotlib.pyplot as plt\n\nvisualizer.plot(layer_index = 6)\nplt.savefig("token_to_token.jpg")\n```\n\n<p align="center">\n    <img alt="token to token" src="images/token_to_token.jpg" />\n</p>\n\n<p align="center">Plot attention matrices normalized on head axis</p>\n\nYou can specify the `order` used in `torch.linalg.norm` in `__call__` and `compute` methods. By default, it\'s a L2 norm.\n\n```python\nfrom transformers_visualizer import TokenToTokenNormalizedAttentions\n\nvisualizer = TokenToTokenNormalizedAttentions(model, tokenizer)\nvisualizer.compute(text).plot()\n```\n\n<p align="center">\n    <img alt="normalized token to token"src="images/token_to_token_normalized.jpg" />\n</p>\n\n</details>\n\n## Upcoming features\n\n- [ ] Adding an option to specify head/layer indices to plot.\n- [ ] Adding other plotting backends such as Plotly, Bokeh, Altair.\n- [ ] Implement other visualizers such as [vector norm](https://arxiv.org/pdf/2004.10102.pdf).\n\n## References\n\n- [[1]][captum_part1] Captum\'s BERT example part 1\n- [[2]][captum_part2] Captum\'s BERT example part 2\n\n\n## Acknowledgements\n\n- [Transformers Interpret](https://github.com/cdpierse/transformers-interpret) for the idea of this project.\n\n[Captum]: https://captum.ai/\n[captum_part1]: https://captum.ai/tutorials/Bert_SQUAD_Interpret\n[Captum_part2]: https://captum.ai/tutorials/Bert_SQUAD_Interpret2',
    'author': 'VDuchauffour',
    'author_email': 'vincent.duchauffour@proton.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/VDuchauffour/transformers-visualizer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import Enum
from typing import List, Optional, Tuple

import matplotlib.pyplot as plt
from torchtyping import TensorType

sequence_input, num_head = None, None


class SubplotColumnsConfig(Enum):
    """
    Number of subplot columns given plot type.
    """

    TOKEN_TO_TOKEN = 3
    TOKEN_TO_HEAD = 2


class SubplotConfig:
    """
    Subplot config given plot type.
    """

    def __init__(self, length_dimension: int, plot_type: SubplotColumnsConfig):
        """
        Instantiate a subplot config. Set `n_cols` and `n_rows` given the type of plot.

        Args:
            length_dimension (int): length of the dimension to plot.
            plot_type (SubplotColumnsConfig): plot type.
        """
        self.plot_type = plot_type.name
        self.n_cols = plot_type.value
        self.n_rows = max(1, length_dimension // self.n_cols)


def plot_token_to_token(
    matrice: TensorType["sequence_input", "sequence_input"],
    tokens: List[str],
    dimension_type: str,
    num_dimension: int,
    ax: Optional[plt.Axes] = None,
    ticks_fontsize: int = 7,
    title_fontsize: int = 9,
    cmap: str = "viridis",
    colorbar: bool = True,
) -> plt.Axes:
    """
    Generates a token-to-token matplotlib plot. A `torch.Tensor` is expected as input.

    Args:
        matrice (torch.Tensor): `torch.Tensor` to plot. Expect tensor of dimension `"sequence_input", "sequence_input"`.
        tokens (List[str]): List of tokens to plot on x and y ticks label.
        dimension_type (str): Dimension name to plot.
        num_dimension (int): Dimension number plotted in title.
        figsize (Tuple[int, int], optional): Figsize of the plot. Defaults to (20, 20).
        ticks_fontsize (int, optional): Ticks fontsize. Defaults to 7.
        title_fontsize (int, optional): Title fontsize. Defaults to 9.
        cmap (str, optional): Colormap. Defaults to "viridis".
        colorbar (bool, optional): Display colorbars. Defaults to True.

    Returns:
        plt.Axes: The generated axes.
    """
    if ax is None:
        ax = plt.gca()
    assert ax is not None

    im = ax.imshow(matrice.detach().cpu().numpy(), cmap=cmap)

    ax.set_xticks(range(len(tokens)))
    ax.set_yticks(range(len(tokens)))
    ax.set_xticklabels(tokens, fontdict={"fontsize": ticks_fontsize}, rotation=90)
    ax.set_yticklabels(tokens, fontdict={"fontsize": ticks_fontsize})
    ax.set_title(
        f"{dimension_type} {num_dimension}",
        fontdict={"fontsize": title_fontsize},
    )
    if colorbar:
        plt.colorbar(im, ax=ax, fraction=0.15, pad=0.05)
    return ax


def plot_token_to_token_specific_dimension(
    matrices: TensorType[..., "sequence_input", "sequence_input"],
    tokens: List[str],
    dimension_name: str,
    figsize: Tuple[int, int] = (20, 20),
    ticks_fontsize: int = 7,
    title_fontsize: int = 9,
    cmap: str = "viridis",
    colorbar: bool = True,
):
    """
    Generates a token-to-token matplotlib plot for all indices of a specific dimension. A `torch.Tensor` is expected as input.

    Args:
        matrices (torch.Tensor): `torch.Tensor` to plot. Expect tensor of dimension `"...", "sequence_input", "sequence_input"`.
        tokens (List[str]): List of tokens to plot on x and y ticks label.
        dimension_name (str): Dimension name to plot.
        figsize (Tuple[int, int], optional): Figsize of the plot. Defaults to (20, 20).
        ticks_fontsize (int, optional): Ticks fontsize. Defaults to 7.
        title_fontsize (int, optional): Title fontsize. Defaults to 9.
        cmap (str, optional): Colormap. Defaults to "viridis".
        colorbar (bool, optional): Display colorbars. Defaults to True.

    Returns:
        matplotlib.figure.Figure: The generated matplotlib figure.
    """
    length_dimension = matrices.size(0)
    subplot_config = SubplotConfig(
        length_dimension, SubplotColumnsConfig.TOKEN_TO_TOKEN
    )

    fig = plt.figure(figsize=figsize)
    for idx, matrice in enumerate(matrices):
        ax = fig.add_subplot(
            subplot_config.n_rows,
            subplot_config.n_cols,
            idx + 1,
        )
        ax = plot_token_to_token(
            matrice,
            tokens,
            dimension_name,
            idx + 1,
            ax,
            ticks_fontsize,
            title_fontsize,
            cmap,
            colorbar,
        )

    plt.tight_layout()
    return plt.gcf()
